"""
Constants values used in the Forum App.
"""

FORUM_DEFAULT_PAGE = 1
FORUM_DEFAULT_PER_PAGE = 20

FORUM_MAX_DEEP_SEARCH_COMMENT_COUNT = 1000

RETIRED_TITLE = "[deleted]"
RETIRED_BODY = "[deleted]"
