# Generated by Django 4.2 on 2024-10-24 13:36

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("forum", "0003_alter_commentthread_title"),
    ]

    operations = [
        migrations.AddField(
            model_name="comment",
            name="author_username",
            field=models.CharField(
                blank=True,
                help_text="Username at time of posting, preserved for historical accuracy",
                max_length=255,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="commentthread",
            name="author_username",
            field=models.CharField(
                blank=True,
                help_text="Username at time of posting, preserved for historical accuracy",
                max_length=255,
                null=True,
            ),
        ),
        migrations.AddField(
            model_name="commentthread",
            name="retired_username",
            field=models.CharField(
                blank=True,
                help_text="Username to display if author account was retired",
                max_length=255,
                null=True,
            ),
        ),
    ]
